#define F_CPU 6000000UL

#include <inttypes.h>
#include <util/delay.h>
#include <avr/io.h>
#include <avr/interrupt.h>
#include <stdio.h>
#include <avr/sleep.h>


/*
        Tiny2313
        +--------------------------------------------+
  reset |  1 PA2/-reset                       VCC 20 | +5 volt
     rx |  2 PD0/rxd            int7/scl/usck/PB7 19 | sck   
     tx |  3 PD1/txd             int6/do/miso/PB6 18 | miso  
  xtal2 |  4 PA1/xtal2           int5/di/mosi/PB5 17 | mosi  
  xtal1 |  5 PA0/xtal1              int4/oc1b/PB4 16 | data5
  data6 |  6 PD2/int0/ckout         int3/oc1a/PB3 15 | data4
  data7 |  7 PD3/int1               int2/oc0a/PB2 14 | data3
  data8 |  8 PD4/t0                 int1/ain1/PB1 13 | data2
  data9 |  9 PD5/t1/oc0b            int0/ain0/PB0 12 | data1
    gnd | 10 GND                          icp/PD6 11 | 
        +--------------------------------------------+
*/ 




// Define the i/o pins that are connected to the charlieplexed
// display.  The suffix B or D makes it easier to keep track
// of the PORT the bit belongs to.
#define CHARLIE_1_B		PB3
#define CHARLIE_2_B		PB1
#define CHARLIE_3_D		PD5
#define CHARLIE_4_D		PD3
#define CHARLIE_5_D		PD4
#define CHARLIE_6_B		PB0
#define CHARLIE_7_B		PB4
#define CHARLIE_8_B		PB2
#define CHARLIE_9_D		PD6


#define MSG_LENGTH	10		// 10 chars message length 

uint8_t rxPtr = 0;			// Receive buffer write pointer
uint8_t txPtr = 0;			// Send buffer read pointer
uint8_t rxBuf[MSG_LENGTH];	// Receive buffer
uint8_t txBuf[MSG_LENGTH];	// Send buffer

uint8_t digits[8];			// The array containing the digits to be shown on the display
uint8_t volatile stopAttract = 0;	// Set to 1 when first packed has arrived


//
// This is the Interrupt Service Routine that need to get called 
// often enough to not show any visible flickering on the displays.
//
ISR (SIG_OVERFLOW0) { 
	static uint8_t currentdigit=0;	// Keeps track of the digit to be shown
	uint8_t segments;				// The bitpattern of the current digit

	TCNT0 = 128;         			// Reset TCNT0 halfway to get a higher freq

	// Get the bit pattern of the current digit
	segments=digits[currentdigit];

	// Set all LED-portpins as inputs so the corresponding
	// segments of the displays will be off
	DDRB &= ~(
			_BV(CHARLIE_1_B) | 
			_BV(CHARLIE_2_B) | 
			_BV(CHARLIE_6_B) | 
			_BV(CHARLIE_7_B) | 
			_BV(CHARLIE_8_B) 
		);

	DDRD &= ~(
			_BV(CHARLIE_3_D) | 
			_BV(CHARLIE_4_D) | 
			_BV(CHARLIE_5_D) |
			_BV(CHARLIE_9_D)
		);


	// Set low level on all LED-port pins. This prepares the segments
	// to be lit if the pin is changed to output at a later stage. The
	// displays ar of common anode-type (common positive) so the segments
	// needs to be sunk to ground to be turned on.
	PORTB &= ~(
			_BV(CHARLIE_1_B) | 
			_BV(CHARLIE_2_B) | 
			_BV(CHARLIE_6_B) | 
			_BV(CHARLIE_7_B) | 
			_BV(CHARLIE_8_B) 
		);

	PORTD &= ~(
			_BV(CHARLIE_9_D) |
			_BV(CHARLIE_3_D) | 
			_BV(CHARLIE_4_D) | 
			_BV(CHARLIE_5_D)
		);


	// Set portpins as output for segments to be lit
	// We just assume that each segment has it's standard
	// place in the connections.  The special case is handled
	// in the switch-statement below.

	if (bit_is_set(segments,0)) DDRB |= _BV(CHARLIE_1_B);
	if (bit_is_set(segments,1)) DDRB |= _BV(CHARLIE_2_B);
	if (bit_is_set(segments,2)) DDRD |= _BV(CHARLIE_3_D);
	if (bit_is_set(segments,3)) DDRD |= _BV(CHARLIE_4_D);
	if (bit_is_set(segments,4)) DDRD |= _BV(CHARLIE_5_D);
	if (bit_is_set(segments,5)) DDRB |= _BV(CHARLIE_6_B);
	if (bit_is_set(segments,6)) DDRB |= _BV(CHARLIE_7_B);
	if (bit_is_set(segments,7)) DDRB |= _BV(CHARLIE_8_B);


	// Here we do the bit-fiddling thats neccesary to charlieplex.
	// Since one of the segments (each different) of each display 
	// is moved to the 9'th connection we need to take care of that.
	//
	// Depending on which digit that are active now we need to handle
	// the situation in its own unique way.
	//
	// The A segment on the first digit is moved from the 1'th line
	// to the 9'th line so be basically to the same thing as in the bunch
	// of tests above, but only test for the A segment and if it's lit
	// we turn on the 9'th line instead of the first line.  
	// We then need to activate the transistor that handles the common 
	// anode for the first digit. The transistor for the first display
	// is connected to the 1'th line (where the A-segment usualy go).
	// so we turn on the output for that pin and set it high.
	//
	// The next time this routine is called we do the same thing with
	// the second digit. But we then check for the B-segment and so on...
	switch (currentdigit) {
		case 0:
			if (segments & 0b00000001) DDRD |= _BV(CHARLIE_9_D);
			DDRB |= _BV(CHARLIE_1_B);
			PORTB |= _BV(CHARLIE_1_B);
			break;
		case 1:
			if (segments & 0b00000010) DDRD |= _BV(CHARLIE_9_D);
			DDRB |= _BV(CHARLIE_2_B);
			PORTB |= _BV(CHARLIE_2_B);
			break;
		case 2:
			if (segments & 0b00000100) DDRD |= _BV(CHARLIE_9_D);
			DDRD |= _BV(CHARLIE_3_D);
			PORTD |= _BV(CHARLIE_3_D);
			break;
		case 3:
			if (segments & 0b00001000) DDRD |= _BV(CHARLIE_9_D);
			DDRD |= _BV(CHARLIE_4_D);
			PORTD |= _BV(CHARLIE_4_D);
			break;
		case 4:
			if (segments & 0b00010000) DDRD |= _BV(CHARLIE_9_D);
			DDRD |= _BV(CHARLIE_5_D);
			PORTD |= _BV(CHARLIE_5_D);
			break;
		case 5:
			if (segments & 0b00100000) DDRD |= _BV(CHARLIE_9_D);
			DDRB |= _BV(CHARLIE_6_B);
			PORTB |= _BV(CHARLIE_6_B);
			break;
		case 6:
			if (segments & 0b01000000) DDRD |= _BV(CHARLIE_9_D);
			DDRB |= _BV(CHARLIE_7_B);
			PORTB |= _BV(CHARLIE_7_B);
			break;
		case 7:
			if (segments & 0b10000000) DDRD |= _BV(CHARLIE_9_D);
			DDRB |= _BV(CHARLIE_8_B);
			PORTB |= _BV(CHARLIE_8_B);
			break;
	}

	// Show next digit at the next interrupt. This can't be done
	// at the top of the ISR since the digit variable is needed
	// in the switch-statement later on.
	currentdigit++;
	if (currentdigit>7) currentdigit=0;
}


//
// This is the interrupt handler for UART transmit buffer empty.
//
// It sends all characters in the txBuf one bu one.  The length
// of the buffer is a fixed value so all messages must be of this
// length.
//
ISR(SIG_USART0_TX) {
	if (txPtr < MSG_LENGTH) {	// Are there more data top send?
		UDR = txBuf[txPtr];		// Yes, send it.
		txPtr++;				// And point to the next character to send
	}
}

//
// This is the interrupt handler for the UART data received.
//
// Collect all incomming characters in the rxBuf -buffer until
// the end_of_message character is detected at the correct position.
// Then check if the message is for this unit.  If it is decode the
// data and put it in the display array.  Othervise pass it on to the
// next display in the chain.
//
ISR(SIG_USART0_RX) {
	unsigned char rx;

	rx = UDR; 		// Get the incomming data

	// If the MSB if the character is not set then it's a start byte
	if (rx < 128) {	
			rxPtr = 0;		// So reset the buffer pointer to zero
	}

	rxBuf[rxPtr++] = rx;	// Store the character in the buffer

	// If we get the entire message then process it.
	if (rxPtr == MSG_LENGTH) {
		stopAttract=1;
		// Check If the message is for us
		if ((rxBuf[0] & 0x0F)== 0) {
			for (rx=0; rx<8; rx++) {
				digits[rx] = (rxBuf[rx+1] & 0x7F);
			}
			if (bit_is_set(rxBuf[0],6)) digits[7] |= 0x80; 
			for (rx=0; rx<7; rx++) {
				if (bit_is_set(rxBuf[MSG_LENGTH-1],rx)) digits[rx] |= 0x80; 
			}		 

		// Not for us, so we just pass it on
		} else {
			rxBuf[0]--; 	// Decrement the unit-counter
			// Copy the entire message to the transmit buffer
			for (rxPtr=0; rxPtr<MSG_LENGTH; rxPtr++) {
				txBuf[rxPtr] = rxBuf[rxPtr];
			}
			txPtr = 1;		// The next character to be sent by the ISR
			UDR = txBuf[0];	// Send the first character manuallu to trigger the ISR
		}
		rxPtr = 0;			// Reset the receive buffer pointer to zero
	}

	// Make sure that the pointer can't point outside the buffer
	if (rxPtr >= MSG_LENGTH) rxPtr = MSG_LENGTH - 1;
}



#define USART_BAUDRATE 9600
#define BAUD_PRESCALE (((F_CPU / (USART_BAUDRATE * 16UL))) - 1) 

//
// Initialize and start the timer responsible for calling
// the update-the-display routine.  The timer should timeout
// often enough for getting a stable non-flickering display.
//
void InitCPU(void) {

	//8 bit timer initialization (timer 0)
	TIMSK = _BV(TOIE0);   			//Enable TCNT0 overflow
	TCNT0 = 0;         				//Reset TCNT0
	TCCR0A = 0;						//Normal overflow
	TCCR0B = _BV(CS00) |_BV(CS01);  //Prescale /64

	// USART 9600 bps, 8N1
   	UCSRC |= (1 << UCSZ0) | (1 << UCSZ1); // Use 8-bit character sizes
	UBRRL = BAUD_PRESCALE; // Load lower 8-bits of the baud rate value into the low byte of the UBRR register
	UBRRH = (BAUD_PRESCALE >> 8); // Load upper 8-bits of the baud rate value into the high byte of the UBRR register
	UCSRB |= (1 << RXEN) 	// Turn on the reception
			|(1 << TXEN)   	// Turn on the transmission 
			|(1 << RXCIE) 	// Enable RX interrupts
			|(1 << TXCIE); 	// Enable TX interrupts

	sei();					// Globally enable interrupts
}



//
// A generic delay subroutine
//
void Delay_ms(int cnt) {
	while (cnt-->0) {
		_delay_ms(1);
	}
}



// 
// The main procedure.  Inititalise the timer used for updating
// the display, and then show some patterns on the display.
//
int main (void) {
	int  i,j;

	InitCPU();

	Delay_ms(500);

	// Scroll three skewed horizontal lines
	for (i=0; i<24; i++) {
		for (j=0; j<8; j++) {
			digits[j]=0;
			if ( ((i-j)-0>0) && ((i-j)-0<8) ) digits[j]|=0x01;
			if ( ((i-j)-4>0) && ((i-j)-4<8) ) digits[j]|=0x40;
			if ( ((i-j)-8>0) && ((i-j)-8<8) ) digits[j]|=0x08;
		}
		Delay_ms(50);
	}

	// Fill & unfill with vertical lines
	for (i=0; i<16; i++) {
		if (i&1) digits[i>>1]|=0x06;
		if (!(i&1)) digits[i>>1]|=0x30;
		Delay_ms(50);
	}
	for (i=0; i<16; i++) {
		if (i&1) digits[i>>1]&=~0x06;
		if (!(i&1)) digits[i>>1]&=~0x30;
		Delay_ms(50);
	}

	Delay_ms(250);

	// Show "u-disp" plus version
	digits[0] = 0x3e;
	digits[1] = 0x40;
	digits[2] = 0x5e;
	digits[3] = 0x04;
	digits[4] = 0x6d;
	digits[5] = 0x73;
	digits[6] = 0x00;
	digits[7] = 0x06;

	Delay_ms(1000);

	// Sweep a dot back and forth until the first command packet arrived
	while (!stopAttract) {
		for (i=0; i<8; i++) {
			for (j=0; j<8; j++) {
				if (i==j) {
					digits[j]=0x80;
				} else {
					digits[j]=0x00;
				}
			}
			Delay_ms(50);
		}
		for (i=6; i>0; i--) {
			for (j=0; j<8; j++) {
				if (i==j) {
					digits[j]=0x80;
				} else {
					digits[j]=0x00;
				}
			}
			Delay_ms(50);
		}
	}

	// Then just no nothing - attract mode is over
	for (;;);





// set_sleep_mode(SLEEP_MODE_PWR_DOWN);
// set_sleep_mode(SLEEP_MODE_STANDBY);
// sleep_mode();

}

